/*;  khstr.cc {\documentstyle{jsrc2latex}\font\tenrm=cmtt10} */

/*;  $$Revision: 1.8 $$ {\rhead{\sl khstr.cc\quad\thepage}} */
/*;  $$Date: 1998/07/20 06:42:45 $$  */
/*;  $$Source: /home/shiiba/KhProgDevelop/KhC++Lib/string/khstr.cc,v $$  */
/*;  $$RCShost: tone.kuciv.kyoto-u.ac.jp $$  */

//  DYNAMIC STRING CLASS -- khstr.cpp
//  Provides a general dynamic string class.
//  Copyright 1991-1994 Scott Robert Ladd. All Rights Reserved.

#include "khstr.h"
#include <iomanip>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <limits.h>
#include <assert.h>
#include "kherror.h"

//-------------------------------------------------------------
//  String class
//-------------------------------------------------------------

// class-global constant initialization
size_t KhString::AllocIncr = 8;

// calculate the allocation size for a string
inline size_t KhString::CalcSiz(size_t needed)
    {
    size_t x = ((needed + AllocIncr) / AllocIncr) * AllocIncr;
    return x;
    }

// constructor
KhString::KhString()
    {
    Len = 0;
    Siz = 0;
    Txt = NULL;
    }

KhString::KhString(const KhString & str)
    {
    Len = str.Len;
    Siz = str.Siz;

    if (str.Txt == NULL)
        Txt = NULL;
    else
        {
        Txt = new char[Siz];

//        if (Txt == NULL)
//            throw DebugString(*this,SE_ALLOC);
	assert(Txt != NULL);

        memcpy(Txt,str.Txt,Len + 1);
        }
    }

KhString::KhString(const char * cstr)
    {
    if ((cstr == NULL) || (cstr[0] == '\x00'))
        {
        Len = 0;
        Siz = 0;
        Txt = NULL;
        }
    else
        {
        Len = strlen(cstr);
        Siz = CalcSiz(Len);

        Txt = new char [Siz];

//        if (Txt == NULL)
//            throw DebugString(*this,SE_ALLOC);
	assert(Txt != NULL);

        memcpy(Txt,cstr,Len + 1);
        }
    }

KhString::KhString(size_t count, char fillCh)
    {
    if (count == 0)
	 	{
		Len = 0;
		Siz = 0;
		Txt = NULL;
		return;
		}

    Siz = CalcSiz(count);
    Len = count;

    Txt = new char[Siz];

//    if (Txt == NULL)
//        throw DebugString(*this,SE_ALLOC);
    assert(Txt != NULL);

    memset(Txt,fillCh,count);

    Txt[count] = '\x00';
    }

// KhString::KhString(strstream & strm)
KhString::KhString(stringstream & strm)
    {
    // terminate stream with a null
    strm << ends;

    // get number of characters in stream
//     Len = strlen(strm.str());
    Len = (strm.str()).length();

    // calculate size
    Siz = CalcSiz(Len);

    // allocate text buffer
    Txt = new char[Siz];

//    if (Txt == NULL)
//        throw DebugString(*this,SE_ALLOC);
    assert(Txt != NULL);

    // copy characters from stream
//     strcpy(Txt,strm.str());
    strcpy(Txt,(strm.str()).c_str());
    }

// ĤΥ󥹥ȥ饯ϤޤѥǤʤΤǡȥ
// Ȥޤ
// @begin{comment-out-by-Shiiba} ---------------------------------
//KhString::KhString(size_t maxsize, const char * format, ... )
//    {
//    // allocate temporary buffer
//    char * buffer = new char[maxsize];
//
////    if (buffer == NULL)
////        throw DebugString(*this,SE_ALLOC);
//    assert(buffer != NULL);
//
//    // initialize argument list
//    va_list args;
//
//    va_start(args,format);
//
//    // format items into buffer based on format
//    #if (MSC_VER >= 0x700)
//        Len = _vsnprintf(buffer,maxsize,format,args);
//
//        if (Len == size_t(-1))
//            Len = maxsize;
//    #else
//        Len = vsprintf(buffer,format,args);
//    #endif
//
//    // end argument list processing
//    va_end(args);
//
//    // calculate required Txt length
//    Siz = CalcSiz(Len);
//
//    // allocate Txt
//    Txt = new char[Siz];
//
////    if (Txt == NULL)
////        throw DebugString(*this,SE_ALLOC);
//    assert(Txt != NULL);
//
//    // duplicate data from buffer
//    strcpy(Txt,buffer);
//
//    // delete buffer
//    delete buffer;
//    }
// @end{comment-out-by-Shiiba} -----------------------------------


// destructor
KhString::~KhString()
    {
    if (Txt != NULL)
        delete [] Txt;
    }

// assignment method
KhString KhString::operator = (const KhString & str)
    {
    // copy length and size
    Len = str.Len;
    Siz = str.Siz;

    // delete existing buffer
    if (Txt != NULL)
        delete [] Txt;

    if (Siz == 0)
        Txt = NULL;
    else
        {
        // allocate a new buffer
        Txt = new char[Siz];

//        if (Txt == NULL)
//            throw DebugString(*this,SE_ALLOC);
	assert(Txt != NULL);

        // copy source text to destination
        memcpy(Txt,str.Txt,Len + 1);
        }

    // outa here
    return *this;
    }

// KhString KhString::operator = (strstream & strm)
KhString KhString::operator = (stringstream & strm)
    {
    // delete existing text buffer
    if (Txt != NULL)
        delete [] Txt;

    // terminate stream with a null
    strm << ends;

    // get number of characters in stream
//     Len = strlen(strm.str());
    Len = (strm.str()).length();

    // calculate size
    Siz = CalcSiz(Len);

    // allocate text buffer
    Txt = new char[Siz];

//    if (Txt == NULL)
//        throw DebugString(*this,SE_ALLOC);
    assert(Txt != NULL);

    // copy characters from stream
//     strcpy(Txt,strm.str());
    strcpy(Txt,(strm.str()).c_str());

    return *this;
    }

// concatenation methods
KhString operator + (const KhString & str1,const KhString & str2)
    {
    KhString result;

    // calculate total length of string
    unsigned long totalLen = (unsigned long)str1.Len
	 						+ (unsigned long)str2.Len;

    // if adding two 0-length strings, return a 0-length string
    if (totalLen == 0)
        return result;

    // make sure that the combined length isn't too long
//    if (totalLen > UINT_MAX)
//        throw StringEx(SE_TOO_LONG);
    assert(!(totalLen > UINT_MAX));

    // set size and length
    result.Len = 0;
    result.Siz = KhString::CalcSiz((size_t)totalLen);

    // allocate buffer
    result.Txt = new char[result.Siz];

//    if (result.Txt == NULL)
//        throw StringEx(SE_ALLOC);
    assert(result.Txt != NULL);

    // plkace a terminating character
    result.Txt[0] = '\000';

    // copy str1.Txt
    if (str1.Txt != NULL)
        {
        memcpy(result.Txt,str1.Txt,str1.Len);
        result.Len = str1.Len;
        }

    // copy str2.Txt
    if (str2.Txt != NULL)
        {
        memcpy(&result.Txt[result.Len],str2.Txt,str2.Len + 1);
        result.Len += str2.Len;
        }

    return result;
    }

// concatenation methods
KhString operator + (const KhString & str, char ch)
    {
    KhString result;

    if (str.Txt == NULL)
        {
        // if the string is empty, it contains one character
        result.Len = 1;
        result.Siz = KhString::AllocIncr;
        result.Txt = new char [result.Siz];

//        if (result.Txt == NULL)
//            throw StringEx(SE_ALLOC);
	assert(result.Txt != NULL);

        result.Txt[0] = ch;
        result.Txt[1] = '\000';
        }
    else
        {
        // if string is maximum length, error
//        if (str.Len == UINT_MAX)
//            throw StringEx(SE_TOO_LONG);
	  assert(str.Len != UINT_MAX);

        // increment length of string
        result.Len = str.Len + 1;

        // if length = size, increase size
        if (result.Len == str.Siz)
            result.Siz = str.Siz + KhString::AllocIncr;
        else
            result.Siz = str.Siz;

        // allocate buffer
        result.Txt = new char[result.Siz];

//        if (result.Txt == NULL)
//            throw StringEx(SE_ALLOC);
	  assert(result.Txt != NULL);

        // copy orginal string
        memcpy(result.Txt,str.Txt,str.Len);

        // append character
        result.Txt[str.Len]  = ch;
        result.Txt[result.Len] = '\000';
        }

    return result;
    }

KhStrCompVal KhString::Compare(const KhString & str,
                           KhStrCompMode caseChk) const
    {
    // handle special cases where one string is empty
    if (Txt == NULL)
        if (str.Txt == NULL)
            return SC_EQUAL;
        else
            return SC_LESS;

    if (str.Txt == NULL)
        return SC_GREATER;

    // compare the # of characters in the shorter string
    size_t count;

    if (str.Len < Len)
        count = str.Len;
    else
        count = Len;

    // working variables
    char   c1, c2;
    size_t i;

    if (caseChk == SM_IGNORE)
        {
        // case insensitive comparison
        for (i = 0; i < count; ++i)
            {
            c1 = (char)tolower(Txt[i]);
            c2 = (char)tolower(str.Txt[i]);

            // if characters differ
            if (c1 != c2)
                {
                // select appropriate result
                if (c1 < c2)
                    return SC_LESS;
                else
                    return SC_GREATER;
                }
            }
        }
    else
        {
        for (i = 0; i < count; ++i)
            {
            c1 = Txt[i];
            c2 = str.Txt[i];

            // if characters differ
            if (c1 != c2)
                {
                // select appropriate result
                if (c1 < c2)
                    return SC_LESS;
                else
                    return SC_GREATER;
                }
            }
        }

    // at this point, no differences were found
    if (Len == str.Len)
        return SC_EQUAL;
    else
        {
        // is lengths differ, shorter string < longer one
        if (Len < str.Len)
            return SC_LESS;
        else
            return SC_GREATER;
        }
    }

// substring search methods
KhBoolean KhString::Find(const KhBoyerMoore & bm, size_t & pos) const
    {
    size_t i, j, patlen;

    // store pattern length locally (it gets used a lot)
    patlen = bm.GetPatternLen();

    // i is the index into the target
    i = patlen;

    while (i <= Len)
        {
        // j is an index into pattern
        j = patlen;

        // while corresponding characters match
        while (bm[j - 1] == Txt[i - 1])
            {
            if (j > 1)
                {
                // move left one character for next comparison
                --j;
                --i;
                }
            else
                {
                // we've reached the beginning of the pattern
                // pattern found!
                pos = i - 1;
                return YES;
                }
            }

        // move target index by delta value of
        // mismatched character
        i += bm.GetDelta(Txt[i - 1]);
        }

    return NO;
    }

KhBoolean KhString::Find(const KhString & str,
                     size_t & pos,
                     KhStrCompMode caseChk) const
    {
    // uses the brute force method
    if (Len < str.Len)
        return NO;

    // duplicate buffers
    char * target = new char[Len + 1];

//    if (target == NULL)
//        throw DebugString(*this,SE_ALLOC);
    assert(target != NULL);

    strcpy(target,Txt);

    char * pattern = new char[str.Len + 1];

//    if (pattern == NULL)
//        throw DebugString(*this,SE_ALLOC);
    assert(pattern != NULL);

    strcpy(pattern,str.Txt);

    // create return value variable
    KhBoolean result;

    // convert to all lowercase if case-insensitive search
    if (caseChk == SM_IGNORE)
        {
//        strlwr(target);
//        strlwr(pattern);
	Strlwr(target);
	Strlwr(pattern);
        }

    // calculate last position in *this where str could be
    size_t end = Len - str.Len;
    size_t p, t;

    // start at the beginning of ttarget
    pos = 0;

    for (;;)
        {
        p = 0;   // beginning of pattern
        t = pos; // beginning of search position in target

        // while characters match
        // and we're not at the end of the strings

        while ((pattern[p] == target[t])
            && (pattern[p] != 0)
            && (target[t]  != 0))
            {
            // move to next character
            ++t;
            ++p;
            }

        // if we've reached the end of pattern
        //     we've found pattern in target

        if (pattern[p] == 0)
            {
            result = YES;
            break;
            }

        // if we've reached the end of target
        // or we've searched far enough
        //     pattern has not been found

        if ((target[t] == 0) || (pos >= end))
            {
            result = NO;
            break;
            }

        // keep looking, starting at the mismatch

        ++pos;
        }

    // delete resultorary buffers
    delete target;
    delete pattern;

    // outa here
    return result;
    }

// substring deletion method
void KhString::Delete(size_t pos, size_t count)
    {
    if (Txt == NULL)
        return;

    size_t newLen, i;

    // error if deleting outside of string
//    if ((pos + count - 1) > Len)
//        throw DebugString(*this,SE_INVALID);
    assert(!((pos + count - 1) > Len));

    // length of new string
    newLen = Len - count;

    if ((Siz - newLen) > AllocIncr)
        {
        // allocation size has changed
        // calculate new size

        Siz = CalcSiz(newLen);

        // create new buffer

        char * result = new char[Siz];

//        if (result == NULL)
//            throw DebugString(*this,SE_ALLOC);
	assert(result != NULL);

        // copy characters into new buffer
        char * tptr = result;

        for (i = 0; i <= Len; ++i)
            {
            // when count is reached, skip deleted characters
            if (i == pos)
                i += count;

            *tptr = Txt[i];
            ++tptr;
            }

        // delete old buffer
        delete [] Txt;

        // assign new buffer
        Txt = result;
        }
    else
        {
        // just "slide" characters down
        for (i = pos; i <= pos + count; ++i)
            Txt[i] = Txt[i + count];
        }

    Len = newLen;
    }

// substring insertion methods
void KhString::Insert(size_t pos, char ch)
    {
//    if (pos > Len)
//        throw DebugString(*this,SE_INVALID);
    assert(!(pos > Len));

    if (Txt == NULL)
        {
        // an empty string == ch
        Len = 1;
        Siz = AllocIncr;

        Txt = new char [Siz];

//        if (Txt == NULL)
//            throw DebugString(*this,SE_ALLOC);
	assert(Txt != NULL);

        Txt[0] = ch;
        Txt[1] = '\000';
        }
    else
        {
        size_t newLen = Len + 1;
        size_t i;

        if (newLen == Siz)
            {
            // need a larger buffer
            Siz += AllocIncr;

            // create resultorary buffer
            char * result = new char[Siz];
            char * tptr = result;

//            if (result == NULL)
//                throw DebugString(*this,SE_ALLOC);
	    assert(result != NULL);

            // copy in old buffer, inserting ch when needed
            for (i = 0; i <= Len; ++i)
                {
                if (i == pos)
                    {
                    *tptr = ch;
                    ++tptr;
                    }

                *tptr = Txt[i];
                ++tptr;
                }

            // delete old buffer
            delete [] Txt;

            // assign new buffer and length
            Txt = result;
            Len = newLen;
            }
        else
            {
            // slide characters right
            for (i = newLen; i > pos; --i)
                Txt[i] = Txt[i-1];

            // insert character
            Txt[pos] = ch;

            // adjust length
            Len = newLen;
            }
        }
    }

void KhString::Insert(size_t pos, const KhString & str)
    {
    if (str.Txt == NULL)
        return;

//    if (pos > Len)
//        throw DebugString(*this,SE_INVALID);
    assert(!(pos > Len));

    if (Txt == NULL)
        {
        // empty string = str
        *this = str;
        }
    else
        {
        // calculate new length
        unsigned long totalLen = str.Len + Len;

//        if (totalLen > UINT_MAX)
//            throw DebugString(*this,SE_TOO_LONG);
	assert(!(totalLen > UINT_MAX));

        size_t i, j;

        // if new  length > current size
        if (totalLen > Siz)
            {
            // allocate new buffer
            Siz = CalcSiz((size_t)totalLen);

            char * result = new char [Siz];
            char * tptr = result;

            // copy buffers from source strings
            for (i = 0; i <= Len; ++i)
                {
                if (i == pos)
                    {
                    for (j = 0; j < str.Len; ++j)
                        {
                        *tptr = str.Txt[j];
                        ++tptr;
                        }
                    }

                *tptr = Txt[i];
                ++tptr;
                }

            // delete old buffer
            delete [] Txt;

            // assign new buffer
            Txt = result;
            }
        else
            {
            // slide section old buffer to right
            for (i = Len + str.Len; i > pos + str.Len; --i)
                Txt[i] = Txt[i - str.Len];

            // insert new string
            for (i = 0; i < str.Len; ++i)
                Txt[pos + i] = str.Txt[i];
            }

        Len = (size_t)totalLen;
        }
    }

// substring retrieval method
KhString KhString::Cut(size_t start, size_t count) const
    {
//    if ((start + count) > Len)
//        throw DebugString(*this,SE_INVALID);
    assert(!((start + count) > Len));

    KhString result;

    if ((start < Len) && (count > 0))
        {
        result.Len = count;
        result.Siz = CalcSiz(count);
        result.Txt = new char[result.Siz];

//        if (result.Txt == NULL)
//            throw DebugString(*this,SE_ALLOC);
	assert(result.Txt != NULL);

        memcpy(result.Txt,&Txt[start],count);

        result.Txt[count] = '\000';
        }

    return result;
    }

KhString KhString::CutHead(size_t count) const
    {
//    if (count > Len)
//        throw DebugString(*this,SE_INVALID);
    assert(!(count > Len));

    KhString result;

    if (count > 0)
        {
        result.Len = count;
        result.Siz = CalcSiz(count);
        result.Txt = new char[result.Siz];

//        if (result.Txt == NULL)
//            throw DebugString(*this,SE_ALLOC);
	assert(result.Txt != NULL);

        memcpy(result.Txt,Txt,count);

        result.Txt[count] = '\000';
        }

    return result;
    }

KhString KhString::CutTail(size_t count) const
    {
//    if (count > Len)
//        throw DebugString(*this,SE_INVALID);
    assert(!(count > Len));

    KhString result;

    if (count > 0)
        {
        result.Len = count;
        result.Siz = CalcSiz(count);
        result.Txt = new char[result.Siz];

//        if (result.Txt == NULL)
//            throw DebugString(*this,SE_ALLOC);
	assert(result.Txt != NULL);

//        memcpy(result.Txt,&Txt[Len - count - 1],count);
	memcpy(result.Txt,&Txt[Len - count],count);

        result.Txt[count] = '\000';
        }

    return result;
    }

// Strupr and Strlwr were added by T.Sakakibara.
char * KhString::Strupr(char * cstr) const
    {
    size_t size;

    size = strlen(cstr);

    size_t i;

    for(i = 0; i < size; ++i)
	cstr[i] = (char)toupper(cstr[i]);

    return cstr;
    }

char * KhString::Strlwr(char * cstr) const
    {
    size_t size;

    size = strlen(cstr);

    size_t i;

    for(i = 0; i < size; ++i)
	cstr[i] = (char)tolower(cstr[i]);

    return cstr;
    }

// case-modification methods
void KhString::ToUpper()
    {
    if (Txt != NULL)
//        strupr(Txt);
	Strupr(Txt);
    }

void KhString::ToLower()
    {
    if (Txt != NULL)
//        strlwr(Txt);
	Strlwr(Txt);
    }

KhString KhString::AsUpper() const
    {
    KhString result = *this;

    if (result.Txt != NULL)
//        strupr(result.Txt);
	Strupr(result.Txt);

    return result;
    }

KhString KhString::AsLower() const
    {
    KhString result = *this;

    if (result.Txt != NULL)
//        strlwr(result.Txt);
	Strlwr(result.Txt);

    return result;
    }

// stream I/O functions

istream& getline(	// ϥȥ꡼फɤ߼
  istream& strm,	// ϥȥ꡼ؤλ
  KhString& str,	// ɤ߼äԤǼ KhStringؤλ
  char delim)		// ǥ᥿ʸ
{
  static const size_t bufsize = 256;
  static char buf[bufsize];

  str = "";
 retry:
  strm.get(buf, bufsize, delim);
  char endline;
  str = str + buf;
  if (strm.get(endline) && endline != delim) {
    // there are some characters not read in
    strm.putback(endline);
    goto retry;
  }
  return strm;
}

istream & operator >> (istream & strm, KhString & str)
{
  static const size_t bufsize = 256;
  static char buf[bufsize];

  if (!strm) {
    buf[0] = '\0';
    str = buf;
    return strm;
  }

  size_t ipos = 0;
  char c;

  while (strm.get(c)) {
    if (!isspace(c)) break;
  }
  if (strm.eof()) {
    buf[0] = '\0';
    str = buf;
    return strm;
  }
  if (c == '\'' || c == '\"') {
    char cmt = c;
    while (strm.get(c)) {
      if (c == cmt) break;
      if (ipos > bufsize - 1) {
	kherror(); abort();
      }
      buf[ipos++] = c;
    }
    assert(c == cmt);
    buf[ipos] = '\0';
  }
  else {
    buf[ipos++] = c;
    while (strm.get(c)) {
      if (isspace(c)) break;
      if (ipos > bufsize - 1) {
	kherror(); abort();
      }
      buf[ipos++] = c;
    }
    if (isspace(c)) {
      strm.putback(c);
    }
    else {
      strm.clear();
    }
    buf[ipos] = '\0';
  }

  str = buf;
  return strm;
}

//-------------------------------------------------------------
//  KhBoyerMoore class
//-------------------------------------------------------------

const size_t KhBoyerMoore::DeltaSize = 256;

KhBoyerMoore::KhBoyerMoore(const KhString & pat)
    : Pattern(pat)
    {
    // allocate delta table
    Delta = new size_t [DeltaSize];

//    if (Delta == NULL)
//        throw KhBoyerMooreEx();
    assert(Delta != NULL);

    // clear table
    size_t i;

    // get length of pattern
    size_t patlen = Pattern.Length();

    for (i = 0; i < DeltaSize; ++i)
        Delta[i] = patlen;

    // set table values
    for (i = 1; i < patlen; ++i)
        Delta[(size_t)Pattern[i - 1]] = patlen - i;

    // set value for last pattern character
    Delta[(size_t)Pattern[patlen - 1]] = 1;
    }

KhBoyerMoore::KhBoyerMoore(const KhBoyerMoore & bm)
    : Pattern(bm.Pattern)
    {
    // allocate delta table
    Delta = new size_t [DeltaSize];

//    if (Delta == NULL)
//        throw KhBoyerMooreEx();
    assert(Delta != NULL);

    // copy contents of source
    memcpy(Delta,bm.Delta,DeltaSize * sizeof(char));
    }

void KhBoyerMoore::operator = (const KhBoyerMoore & bm)
    {
    Delta = bm.Delta;
    }

int read_int(const char * prompt)
{
  int a;
  KhString line;
  cerr << prompt;

  if (!getline(cin, line) || line.Length() == 0) {
    kherror(); abort();
  }

  //  istrstream is(line); // mod. by iYutaka (Tue Sep  9 12:34:26 2003)
//   istrstream is((const char *)line); // mod. by iYutaka (Wed Nov 24 10:24:19 2004)
  istringstream is((const char *)line);

  is >> a;
  return a;
}

double read_double(const char * prompt)
{
  double a;
  KhString line;
  cerr << prompt;

  if (!getline(cin, line) || line.Length() == 0) {
    kherror(); abort();
  }
  //  istrstream is(line);
//   istrstream is((const char *)line);
  istringstream is((const char *)line);
  is >> a;
  return a;
}

KhString read_string(const char * prompt)
{
  KhString a;
  KhString line;
  cerr << prompt;

  if (!getline(cin, line)) {
    kherror(); abort();
  }
  //  istrstream is(line);
//   istrstream is((const char *)line);
  istringstream is((const char *)line);
  is >> a;
  return a;
}

#ifdef _UNIX

void
ttyin(		  // /dev/tty ʸɤߤ
  KhString& line  // ɤߤʸ뤿ʸ
  )
{
  static const size_t bufsize = 256;
  static char buf[bufsize];
  static FILE *tty = NULL;

  if (tty == NULL) {
    tty = fopen("/dev/tty", "r");
  }
  if (fgets(buf, bufsize, tty) == NULL) {
    exit(0);
  }
  else {
    buf[strlen(buf)-1] = '\0';
    line = buf;
  }
}

#endif /* _UNIX */

/*; end of file */
